#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif
#ifdef HAVE_PWD_H
#include <pwd.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif

#include <sha1.h>

#include "tinyaes/tinyaes.h"

#define BUF_SIZE 16 * 63  /* 16 blocks */

extern char *optarg;
extern int optind;

void usage()
{
	printf("aes-crypt: [d] [-l keylen] [-k keyfile] [-p passwd] infile\n");
	exit(1);
}

/* simple iv creater, from sha1[0-16,0-16] */
void
iv_create(char pass[MAXKB], char iv[MAX_IV_SIZE])
{
	int i;
	SHA1_CTX sha1_ctx;
	u8 results[SHA1_DIGEST_LENGTH];

	SHA1Init(&sha1_ctx);
	SHA1Update(&sha1_ctx, (u8 *)pass, MAXKB / 2);
	SHA1Final(results, &sha1_ctx);
	memcpy(iv, results, MAX_IV_SIZE / 2);

	SHA1Init(&sha1_ctx);
	SHA1Update(&sha1_ctx, (u8 *)pass + MAXKB / 2 - 1, MAXKB / 2);
	SHA1Final(results, &sha1_ctx);
	memcpy(iv + MAX_IV_SIZE / 2 - 1, results, MAX_IV_SIZE / 2);
	return;
}

#define iv_copy(d, s) memcpy(d, s, MAX_IV_SIZE)

int
main(int argc, char *argv[])
{
	aes_ctx *ctx;
	unsigned char buf1[BUF_SIZE + 16], buf2[BUF_SIZE + 16];
	static unsigned char pass[MAXKB];
	unsigned char iv_orig[MAX_IV_SIZE], iv_work[MAX_IV_SIZE];
	int mode = 1;
	FILE *ifp = stdin;
	FILE *ofp = stdout;
	FILE *kfp = NULL;
	int have_pass = 0;
	int ch;
	int keylen = 128;
	int ilen, olen;
	int readbuf;
	int blocks = 0;
 
	while ((ch = getopt(argc, argv, "dp:k:l:o:")) != -1) {
		switch (ch) {
		case 'd':
			mode = 0; /* decrypt */
			break;
		case 'p':
			strncpy(pass, optarg, sizeof(pass));
			pass[MAXKB - 1] = 0;
			have_pass = 1;
			break;
		case 'k':
			if ((kfp = fopen(optarg, "rb")) == NULL) {
				(void)fprintf(stderr,
					      "aes-crypt: key file open error %s\n", optarg, strerror(errno));
				exit(1);
			}
			fseek(kfp, 0L, SEEK_END);
			ilen = ftell(kfp);
			fseek(kfp, 0L, SEEK_SET);
			if (MAXKB < ilen) {
				fread(pass, 1, MAXKB, kfp);
				pass[MAXKB - 1] = 0;
			} else {
				fread(pass, 1, ilen, kfp);
				pass[ilen - 1] = 0;
			}
			fclose(kfp);
			have_pass = 1;
			break;
		case 'l':
			keylen = atoi(optarg);
			break;
		case 'o':
			if ((ofp = fopen(optarg, "wb")) == NULL) {
				(void)fprintf(stderr,
					      "aes-crypt: file open error %s\n", optarg, strerror(errno));
				exit(1);
			}
			break;
		case '?':
		default:
			usage();
		}
	}

	if (argc != 2)
		usage();

	if (strcmp(argv[1], "-") != 0) {
		if ((ifp = fopen(argv[1], "rb")) == NULL) {
			(void)fprintf(stderr,
				      "aes-crypt: file open error %s\n", argv[1], strerror(errno));
			exit(1);
		}
	}

	if (!have_pass) {
#ifdef HAVE_GETPASS
		strncpy(pass, getpass("Password: "), sizeof(pass));
		pass[MAXKB - 1] = 0;
#else
		usage();
		exit(1);
#endif
	}
	if (keylen != 128 && keylen != 192 && keylen != 256) {
		fprintf(stderr, "invalid key length (128, 192, 256)\n");
		exit(1);
	}

	if (mode)
		readbuf = BUF_SIZE;
	else
		readbuf = BUF_SIZE + 16;

	iv_create(pass, iv_orig);
	iv_copy(iv_work, iv_orig);
	ctx = aes_init_key(pass, keylen);

	while ((ilen = fread(buf1, 1, readbuf, ifp)) == readbuf) {
		if (mode)
			olen = aes_cbc_encrypt(ctx, &iv_work, buf2, buf1, ilen);
		else
			olen = aes_cbc_decrypt(ctx, &iv_work, buf2, buf1, ilen);
#if 0
		fprintf(stderr, "ilen:%d olen:%d\n", ilen, olen);
#endif
		if (olen != -1)
			fwrite(buf2, 1, olen, ofp);
		else {
			fprintf(stderr, "unrecoverable error %d:%d\n", ilen, olen);
			exit(1);
		}
		blocks++;
	}
	/* last block */
	if (ilen != 0) {
		if (mode)
			olen = aes_cbc_encrypt_finish(ctx, &iv_work, buf2, buf1, ilen);
		else
			olen = aes_cbc_decrypt_finish(ctx, &iv_work, buf2, buf1, ilen);
		if (olen != -1)
			fwrite(buf2, 1, olen, ofp);
		else {
			fprintf(stderr, "unrecoverable error %d:%d\n", ilen, olen);
			exit(1);
		}
		blocks++;
	}
	aes_delete_key(ctx);
	if (mode)
		fprintf(stderr, "%d blocks encrypted\n", blocks);
	else
		fprintf(stderr, "%d blocks decrypted\n", blocks);
	return 0;
}

